### Copyright (C) 2006-2007 Manuel Ospina <mospina@redhat.com>

# This file is part of miaow.
#
# miaow is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# miaow is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with miaow; if not, write to the Free Software
# Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA

"""Localization Data 

This module contains and manages the localization data of one or more files.
These files can be separate documents or files contained in XLIFF document.
"""

import os, os.path
import xliff
import poParser

class TU:
    
    """Container for a translation unit data.
    
    altTrans is a list of dictionaries. Each dictionary has the following keys:
        - stext: (str)
        - target: (str)
        - quality: (int)
    """
    
    # [TODO]
    # To avoid long assigments, create a constructor.
    def __init__(self):
        self.id = None
        self.source = r""
        self.target = r""
        self.state = ""
        self.altTrans = []

class L10nFile:

    """Container for the localization file data."""
    
    def __init__(self):
        self.filepath = ""
        self.slang = ""
        self.tlang = ""
        self.datatype = ""
        self.original = ""
        self.tu = []

class L10nData:
    
    """Handle the localization data in general.

    Usage:
      L10nObj = L10nData(filename=None, format="xliff")

      while L10nObj.next():
        print L10nObj.tu.source
        print L10nObj.tu.target

    Methods:
        next() - Go to next translation unit.
        prev() - Go to previous translation unit
        add_file(L10nFile) - add a new file to the data.
        add_tu(TU) - add a new translation unit to the current file.
        write(filename, format="xliff") - Save the current data.
    """
    
    def __init__(self, filename=None, format="xliff"):
        """Read data from filename. Initialize cursor (tu).
        
        format - supported formats are:
            * xliff
            * po
        """
        self.filename = filename
        self.format = format
        self.cFile = 0
        self.cTu = 0
        self.l10nFiles = []
        self.l10nFile = None
        self.tu = None
        if self.filename:
            if not os.path.isfile(self.filename):
                raise IOError, "File not found"
            # [TODO]
            # A dispatcher will we more maintainable:
            #   meth = get_attr(format)
            #   CALL meth
            if self.format == "xliff":
                self._read_xliff()
            elif self.format == "po":
                self._read_po()
            self.cFile = 0
            self.cTu = 0
            self._update_cursor()
                
    def add_file(self, l10nfile):
        """Add a new file to the localization data."""
        if not isinstance(l10nfile, L10nFile):
            raise TypeError, "argument must be a L10nFile object"
        self.l10nFiles.append(l10nfile)
        self.cFile = len(self.l10nFiles) - 1

    def add_tu(self, l10ntu):
        """Add a new tu to the current file."""
        if not isinstance(l10ntu, TU):
            raise TypeError, "argument must be a TU object."""
        self.l10nFiles[self.cFile].tu.append(l10ntu)
        self.cTu = len(self.l10nFiles[self.cFile].tu) - 1
        self._update_cursor()

    def _update_cursor(self):
        """Update self.tu and self.l10nFile."""
        self.l10nFile = self.l10nFiles[self.cFile]
        self.tu = self.l10nFiles[self.cFile].tu[self.cTu]

    def next(self):
        """Go to next translation unit (tu)."""
        
        lastTu = len(self.l10nFile.tu) - 1
        lastFile = len(self.l10nFiles) - 1
        if self.cTu == lastTu:
            if self.cFile == lastFile:
                return
            else:
                self.cFile += 1
                self.cTu = 0
        else:
            self.cTu += 1
        assert self.cFile < len(self.l10nFiles)
        assert self.cTu < len(self.l10nFiles[self.cFile].tu)
        self._update_cursor()

    def prev(self):
        """Go to previous translation unit (tu)."""
        
        if self.cTu == 0:
            if self.cFile == 0:
                return
            else:
                self.cFile -= 1
                self.cTu = 0
        else:
            self.cTu -= 1
        assert self.cFile >= 0
        assert self.cTu >= 0
        self._update_cursor()
    
    def search_source(self, stext):
        """Return the TU object that contains stext."""
        # [NOTE]
        # This just return the first tu found. This is not enough.
        
        for xfile in self.l10nFiles:
            for xtu in xfile.tu:
                if xtu.source == stext:
                    return xtu
        return None
        
    def _read_xliff(self):
        """Extract data from a XLIFF file."""
         
        xliffObj = xliff.ReadXLIFF(self.filename)
        pFile = None
        pTu = None
        while True:
            cTu = xliffObj.transunit
            if cTu == pTu:
                break
            assert cTu != pTu
            cFile = xliffObj.fileNode
            if cFile != pFile:
                tmpL10n = L10nFile()
                # [NOTE]
                # Although L10nData tries this attributes as optional, the XLIFF
                # specification does not. Only target-language is optional. The
                # User interface should check that those attributes are filled.
                try:
                    tmpL10n.datatype = cFile.getAttribute('datatype')
                except:
                    pass
                try:
                    tmpL10n.original = cFile.getAttribute('original')
                except:
                    pass
                try:
                    tmpL10n.slang = cFile.getAttribute('source-language')
                except:
                    pass
                try:
                    tmpL10n.tlang = cFile.getAttribute('target-language')
                except:
                    pass
                self.add_file(tmpL10n)
            tmpTu = TU()
            tmpTu.id = cTu.getAttribute('id')
            tmpTu.source = xliffObj.read_source()
            tmpTu.target = xliffObj.read_target()
            if tmpTu.target:
                try:
                    tmpTu.state = xliffObj.transunit.getAttribute('state')
                except:
                    pass
            else:
                tmpTu.state = 'needs-translation'
            tmpTu.altTrans = xliffObj.read_altTrans()
            self.add_tu(tmpTu)
            pFile = cFile
            assert pFile != None
            pTu = cTu
            assert pTu != None
            xliffObj.next_transunit()
    
    def _read_po(self):
        """Read l10nData from a po/pot file."""

        fd = open(self.filename)
        poObj = poParser.Po(fd)
        fd.close()
        # [NOTE]
        # As the po file doesn't have enough metadata, _read_po() does not try
        # to guess any value. Users should fill missed information.
        tmpFile = L10nFile()
        tmpFile.filepath = os.path.abspath(self.filename)
        tmpFile.slang = "en" 
        self.add_file(tmpFile)
        i = 0
        for entry in poObj.msg[1:]:
            i += 1
            # [TODO]
            # There is still no way to add the comments from the po file. TU
            # needs a "note" member. XLIFF accepts note tags. 
            tmpTU = TU()
            tmpTU.source = entry['msgid']
            tmpTU.target = entry['msgstr']
            tmpTU.state = entry['status']
            tmpTU.id = i
            self.add_tu(tmpTU)

    def write(self, format='xliff'):
        """Write data into filename."""
        # [TODO]
        # a dispatcher will be much better. See "Diving into python",
        # the chapter about Introspection.
        
        if format == 'po':
            return self._write_po()
        #elif:
        else:
            return self._write_xliff()
    
    def _write_xliff(self):
        """Return data in XLIFF format."""

        if self.filename and self.format == "xliff":
            return self._edit_xliff()
        xliffObj = xliff.WriteXLIFF()
        for xfile in self.l10nFiles:
            xliffObj.write_file(xfile.datatype, xfile.original, \
                                xfile.slang, xfile.tlang)
            for xtu in xfile.tu:
                xliffObj.write_transunit(xtu.id, xtu.source, xtu.target, \
                                         xtu.state, xtu.altTrans)
        return xliffObj.docXLIFF.toxml()
        
    def _write_po(self):
        """Return data in POT format."""
        
        if self.filename and self.format == "po":
            return self._edit_po()
        fd = os.tmpfile()
        fd.write('msgid ""\n')
        fd.write('msgstr ""\n')
        fd.write('\n')
        for xfile in self.l10nFiles:
            for xtu in xfile.tu:
                if xtu.state in ('fuzzy', 'needs-review-translation'):
                    fd.write('#, fuzzy\n')
                fd.write('msgid "%s"\n' % xtu.source)
                fd.write('msgstr "%s"\n\n' % xtu.target)
        fd.seek(0,0)
        ret = fd.read()
        fd.close()
        return ret

    def _edit_xliff(self):
        """Return data from an existing XLIFF file in XLIFF format."""
        
        assert os.path.isfile(self.filename)
        wTu = self.cTu
        wFile = self.cFile
        self.cTu = self.cFile = 0
        self._update_cursor()
        xliffObj = xliff.ReadXLIFF(self.filename)
        prevNode = None
        while True:
            # {Invariant: prevNode != curNode}
            assert prevNode != xliffObj.transunit
            xtu = self.search_source(xliffObj.read_source())
            if xtu:
                xliffObj.edit_target(xtu.target, xtu.state, xtu.altTrans)
            self.next()
            prevNode = xliffObj.transunit
            xliffObj.next_transunit()
            if prevNode == xliffObj.transunit:
                break
        self.cTu = wTu
        self.cFile = wFile
        self._update_cursor()
        return xliffObj.docXLIFF.toxml()

    def _edit_po(self):
        """Save the data into an existing po file."""
        
        assert os.path.isfile(self.filename)
        wTu = self.cTu
        wFile = self.cFile
        self.cTu = self.cFile = 0
        self._update_cursor()
        fd = open(self.filename)
        poObj = poParser.Po(fd)
        fd.close()
        for entry in poObj.msg[1:]:
            tu = self.search_source(entry['msgid'])
            if tu.state in ('fuzzy', 'needs-review-translation'):
                entry['status'] = 'fuzzy'
            entry['target'] = tu.target
            self.next()
        self.cTu = wTu
        self.cFile = wFile
        self._update_cursor()
        return poObj.write()

# [TEST]
# [TODO]
# Write a test case for _edit_*() methods.
TESTXLIFF = """<?xml version="1.0"?>
   <xliff version="1.1">
    <file source-language="en" datatype="plaintext" 
          original="test" target-language="es">
     <body>
      <trans-unit id="1">
       <source>First they ignore you.</source>
       <target state="final">Primero te ignoran</target>
      </trans-unit>
      <trans-unit id="2">
       <source>Then they laugh at you.</source>
       <alt-trans match-quality="80%">
        <source>they laugh at me</source>
        <target>Se rieron de me</target>
       </alt-trans> 
      </trans-unit>
      <trans-unit id="3">
       <source>Then they fight you</source>
      </trans-unit>
      <trans-unit id="4">
       <source>Then you win.</source>
      </trans-unit>
      <trans-unit id="5">
       <source>Mohandas Gandhi</source>
      </trans-unit>
     </body>
    </file>
   </xliff>
"""
TESTPO ="""#

msgid ""
msgstr ""

msgid "First they ignore you."
msgstr "Primero te ignoran"

msgid "Then they laugh at you."
msgstr ""

msgid "Then they fight you"
msgstr ""

msgid "Then you win."
msgstr ""

msgid "Mohandas Gandhi"
msgstr ""
"""
import unittest
class ConfigTest(unittest.TestCase):
    def setUp(self):
        self.path = "test.miau"
        self.po = "test-l10ndata.po"
        self.slang = "en"
        self.tlang = "es"
        self.text1 = "First they ignore you."
        self.text2 = "Then they laugh at you."
        self.text3 = "Then they fight you."
        self.text4 = "Then you win."
        self.text5 = "Mohandas Gandhi"

    def tearDown(self):
        if os.path.isfile(self.path):
            os.unlink(self.path)
        if os.path.isfile(self.po):
            os.unlink(self.po)
    
    def test_01(self):
        """1. Testing write_xliff."""
        x = L10nData()
        # Adding file:
        f = L10nFile()
        f.datatype = "plaintext"
        f.original = "sample text"
        f.slang = "en-US"
        x.add_file(f)
        # Adding tu:
        t = TU()
        t.id = '1'
        t.source = self.text1
        x.add_tu(t)
        # Adding tu:
        t = TU()
        t.id = '2'
        t.source = self.text2
        x.add_tu(t)
        # Adding tu:
        t = TU()
        t.id = '3'
        t.source = self.text3
        x.add_tu(t)
        # Adding tu:
        t = TU()
        t.id = '4'
        t.source = self.text4
        x.add_tu(t)
        # Adding tu:
        t = TU()
        t.id = '5'
        t.source = self.text5
        x.add_tu(t)
        # Writting file
        fd = open(self.path, "w")
        fd.write(x.write())
        fd.close()
        self.failUnless(os.path.isfile(self.path), 'File was not created.')

    def test_2(self):
        """2. Testing read XLIFF."""
        fd = open(self.path, "w")
        fd.write(TESTXLIFF)
        fd.close()
        o = L10nData(self.path)
        self.failUnless(o.filename == self.path, 'File was not read.')
        self.failUnless(len(o.l10nFiles) == 1, 'Files=%d.' % len(o.l10nFiles))
        prevTu = None
        while True:
            self.failUnless(len(o.tu.source) > 0, 'Found a incorrect source.')
            prevTu = o.cTu
            o.next()
            if prevTu == o.cTu:
                break

    def test_3(self):
        """3. Testing write po."""
        fd = open(self.path, "w")
        fd.write(TESTXLIFF)
        fd.close()
        o = L10nData(self.path)
        self.failUnless(o.filename == self.path, 'File was not read.')
        self.failUnless(len(o.l10nFiles) == 1, 'Files=%d.' % len(o.l10nFiles))
        fd = open(self.po, "w")
        fd.write(o.write("po"))
        fd.close()
        self.failUnless(os.path.isfile(self.po), 'File was not created.')
 
    def test_4(self):
        """4. Testing read PO."""
        fd = open(self.po, "w")
        fd.write(TESTPO)
        fd.close()
        o = L10nData(self.po, "po")
        self.failUnless(o.filename == self.po, 'File was not read.')
        self.failUnless(len(o.l10nFiles) == 1, 'Files=%d.' % len(o.l10nFiles))
        prevTu = None
        while True:
            self.failUnless(len(o.tu.source) > 0, 'Found a incorrect source.')
            prevTu = o.cTu
            o.next()
            if prevTu == o.cTu:
                break

if __name__ == '__main__':
    unittest.main()
