### Copyright (C) 2006-2007 Manuel Ospina <mospina@redhat.com>

# This file is part of miaow.
#
# miaow is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# miaow is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with miaow; if not, write to the Free Software
# Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA

"""Po/Pot parser

A simple parser to read po/pot files.
"""

import os
import re
import textwrap

# [BUG]
# When the PO file starts with the first msgid and not with a space or a 
# header, the header is not created. 
class Po:

    """
    Parse po files

    Usage:
    
        fd = open(fileName)
        po = Po(fd)
        
        for entry in po.msg[1:]:
            msgid = entry['msgid']
            msgstr = entry['msgstr']
            status = entry['status']

        Header = po.msg[0]
            
    msg is a list of dictionaries with the following keys:
        entry = {'comm': "Any comentary",
                 'status': "fuzzy|untranslated" or None,
                 'msgid': "the source language entry",
                 'msgstr': "the target language translation" or None
                 }
    """
    

    def __init__(self, fileDescriptor):
        """Initialize po object."""
        self.msg = []
        self.fd = fileDescriptor
        self._parse()
            
    def _parse(self):
        """Parse the po file."""
        
        # [NOTE: general structure of a po file]
        # -----------------------------------------
        #  white-space
        #  #  translator-comments
        #  #. automatic-comments
        #  #: reference...
        #  #, flag...
        #  msgid untranslated-string
        #  msgstr translated-string
        # -----------------------------------------
        
        REid = re.compile(r'^msgid.+')
        REstr = re.compile(r'^msgstr.+')
        REfuzzy = re.compile(r'^#,\s+fuzzy')
        REcomm = re.compile(r'^#')
        REblank = re.compile(r'^$')
        flag = ''
        msgid = ''
        msgstr = ''
        comm = ''
        status = '' 
        
        # [NOTE]
        # It is possible to ignore all blank lines.
        for line in self.fd.readlines():
            # Identifying section
            if REfuzzy.match(line):
                if flag == 'msgstr':
                    # There is no blank space, no comment.  
                    self.add_msg(msgid, msgstr, comm, status)        
                    comm = status = msgid = msgstr = ''
                status = 'fuzzy'
                continue
            elif REcomm.match(line):
                if flag == 'msgstr':
                    # There is no blank space.  
                    self.add_msg(msgid, msgstr, comm, status)        
                    comm = status = msgid = msgstr = ''
                flag = 'comm'               
            elif REid.match(line):
                if flag == 'msgstr':
                    # There is no blank space nor a comm.
                    self.add_msg(msgid, msgstr, comm, status)        
                    comm = status = msgid = msgstr = ''
                flag = 'msgid'
            elif REstr.match(line):
                # There is always a msgid before msgstr 
                flag = 'msgstr'
            elif REblank.match(line):
                if flag == 'blank':
                    # Two blank lines... ignore
                    continue
                self.add_msg(msgid, msgstr, comm, status)        
                comm = status = msgid = msgstr = ''
                flag = 'blank'
            
            # Getting the data  
            if flag == 'msgid':
                op1 = re.match(r'^msgid\s+"(.+)"$', line)
                op2 = re.match(r'^"(.+)"$', line)
                if op1:
                    msgid = msgid + op1.group(1)
                if op2:
                    msgid = msgid + op2.group(1)
            if flag == 'msgstr':
                op1 = re.match(r'^msgstr\s+"(.+)"$', line)
                op2 = re.match(r'^"(.+)"$', line)
                if op1:
                    msgstr = msgstr + op1.group(1)
                if op2:
                    msgstr = msgstr + op2.group(1)
            if flag == 'comm':
                comm = '%s%s' % (comm, line)
        
        if msgstr or msgid:
            # There is no blank line after last entry.
            self.add_msg(msgid, msgstr, comm, status)        
            comm = status = msgid = msgstr = ''
    
    def add_msg(self, msgid, msgstr=None, comm=None, status=None):
        """Append a msg to the message dictionary."""

        if not msgid and len(self.msg):
            return
        if not msgstr:
            status = 'untranslated'
        entry = {'comm': comm.lstrip(),
                 'status': status,
                 'msgid': msgid.lstrip(),
                 'msgstr': msgstr.lstrip()
                 }
        self.msg.append(entry)

    def write(self):
        """Return a string with the content of the po file."""
        
        def write_header(msg):
            ret = ''
            header = 'msgstr ""' 
            tmp = msg['msgstr'].split('\\n')
            for line in tmp:
                if not line:
                    continue
                header = '%s\n"%s\\n"' % (header, line.lstrip())
                
            if 'comm' in msg:
                ret = '%s' % msg['comm'] 
            ret = '%smsgid ""\n%s\n\n' % (ret, header)
            return ret
                   
        fd = os.tmpfile()
        fd.write(write_header(self.msg[0]))
        for entry in self.msg[1:]:
            if 'comm' in entry:
                fd.write(entry['comm'])
            if 'status' in entry:
                if entry['status'] == 'fuzzy':
                    fd.write('#, fuzzy\n')
            fd.write('msgid "%s"\n' % entry['msgid'])
            fd.write('msgstr "%s"\n\n' % entry['msgstr'])

        fd.seek(0,0)    
        ret = fd.read()
        fd.close()
        return ret.rstrip()

# [TEST]
TESTPO ="""
msgid "First they ignore you."
msgstr "Primero te ignoran"

msgid "Then they laugh at you."
msgstr ""

msgid "Then they fight you"
msgstr ""

msgid "Then you win."
msgstr ""

msgid "Mohandas Gandhi"
msgstr ""
"""
import unittest
class ConfigTest(unittest.TestCase):
    def setUp(self):
        self.po = "test-l10ndata.po"

    def tearDown(self):
        if os.path.isfile(self.po):
            os.unlink(self.po)
    
    def test_01(self):
        """1. Testing parser."""
        fd = open(self.po, "w")
        fd.write(TESTPO)
        fd.close()
        p = Po(open(self.po))
        self.failUnless(len(p.msg[0]['msgid']) == 0, 'No header present.')
        for e in p.msg[1:]:
            self.failUnless(len(e['msgid']) > 0, 'Incorrect msgid.')

if __name__ == '__main__':
    unittest.main()
