### Copyright (C) 2006-2007 Manuel Ospina <mospina@redhat.com>

# This file is part of miaow.
#
# miaow is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# miaow is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with miaow; if not, write to the Free Software
# Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA

"""Xliff handler class

This module define the xliff handler class.
"""

# [NOTES]
# ------------------------------------------
# The module should be able to render the following file:
#   <?xml version="1.0"?>
#   <xliff version="1.1">
#    <file source-language="en" datatype="plaintext" 
#          original="test" target-language="es">
#     <body>
#      <trans-unit id="1">
#       <source>First they ignore you.</source>
#       <target state="final">Primero te ignoran</target>
#      </trans-unit>
#      <trans-unit id="2">
#       <source>Then they laugh at you.</source>
#       <alt-trans match-quality="80%">
#        <source>they laugh at me</source>
#        <target>Se rieron de me</target>
#       </alt-trans> 
#      </trans-unit>
#      <trans-unit id="3">
#       <source>Then they fight you</source>
#      </trans-unit>
#      <trans-unit id="4">
#       <source>Then you win.</source>
#      </trans-unit>
#      <trans-unit id="5">
#       <source>Mohandas Gandhi</source>
#      </trans-unit>
#     </body>
#    </file>
#   </xliff>
#
# List of states: new, needs-translation ,needs-review-translation, translated ,final
#
# Read related methods look for this attributes (target-language and state), but
# it do not assume that they exist (They are optional in the XLIFF specification).
# ------------------------------------------
import sys
import os, os.path
from xml.dom.minidom import parse, Document, getDOMImplementation
import xml.dom.minidom

class WriteXLIFF:

    """
    Write XLIFF elements and attributes
    
    Usage:

      obj = WriteXLIFF()
      obj.write_file(datatype, original, slang, tlang=None)
      obj.write_transunit(id, stext, ttext=None, state=None, altTrans)
      fh = open(filename, "w")
      fh.write(obj.docXLIFF.toxml())
      fh.close()
      
    original - the full path to the original document.
    slang - source language
    tlang - target language (a template has not got a target language)
    stext - source text
    ttext - target text
    state - current state of the translation: new, needs-translation, 
            needs-review-translation, translated and final
    altTrans - a list of dictionaries containig a source, a target and an
               optional quality keys
    """
    
    docXLIFF = None
    xliffRoot = None
    bodyNode = None
    
    def __init__(self):
        """Initialize object creating a new document."""
        impl = getDOMImplementation()
        self.docXLIFF = impl.createDocument(None, "xliff", None)
        self.xliffRoot = self.docXLIFF.documentElement

    def write_file(self, datatype, original, slang, tlang=None):
        """Write a new file and its body element. Append them to the root."""
        
        #Create <file>
        dom = Document()
        fileNode = dom.createElement("file")
        fileNode.setAttribute("source-language", slang)
        if tlang:
            fileNode.setAttribute("target-language", tlang)
        fileNode.setAttribute("datatype", datatype)
        fileNode.setAttribute("original", original)
        self.bodyNode = self._write_body()
        fileNode.appendChild(self.bodyNode)
        self.xliffRoot.appendChild(fileNode)

    def _write_body(self):
        """Write a body element."""
        dom = Document()
        bodyNode = dom.createElement("body")
        return bodyNode
    
    def write_transunit(self, id, stext, ttext=None, state=None, altTrans=None):
        """Write a trans-unit element and its source child."""
        assert self.bodyNode != None
        dom = Document()
        tunitNode = dom.createElement("trans-unit")
        tunitNode.setAttribute("id", str(id))
        # Adding <source> subtree:
        sourceDoc = xml.dom.minidom.parseString(r"<source>%s</source>" % stext)
        sourceNode = sourceDoc.documentElement
        tunitNode.appendChild(sourceNode)
        if ttext:
            # Adding <target> subtree
            targetNode = self.write_target(ttext, state)
            tunitNode.appendChild(targetNode)
        if altTrans:
            altTransNode = self.write_altTrans(altTrans)
            for alt in altTransNode:
                tunitNode.appendChild(alt)
        self.bodyNode.appendChild(tunitNode)
    
    def write_target(self, ttext, state=None):
        """Return a new target element."""
        
        tDoc = xml.dom.minidom.parseString(r"<target>%s</target>" % ttext)
        targetNode = tDoc.documentElement
        if state:
            targetNode.setAttribute("state", state)
        return targetNode
        
    def write_altTrans(self, altTrans):
        """Return a list of new alt-trans nodes.
        
        altTrans - a list of dictionaries with a source, target, quality keys.
        target is the only compulsary key.      
        """
        ret = []
        for alt in altTrans:
            if not 'target' in alt:
                # Instead of reporting an error, just ignore this entry
                continue
            dom = Document()
            altTransNode = dom.createElement("alt-trans")
            if 'quality' in alt:
                altTransNode.setAttribute("match-quality", alt['quality'])
            if 'stext' in alt:
                # Adding <source> subtree:
                sDoc = xml.dom.minidom.parseString(r"<source>%s</source>" % \
                                                     alt['stext'])
                sourceNode = sDoc.documentElement
                altTransNode.appendChild(sourceNode)
            targetNode = self.write_target(alt['target'])
            altTransNode.appendChild(targetNode)
            ret.append(altTransNode)
        return ret
        
class ReadXLIFF(WriteXLIFF):

    """
    Read XLIFF elements and attributes.
    
    Usage:
    
      obj = ReadXLIFF(filename)
      prevNode = None
      while prevNode != obj.transunit
        source = obj.read_source()
        target = obj.read_target()
        prevNode = obj.transunit
        obj.next_transunit() 
    
    Other methods:
        obj.prev_transunit()
    """
    
    def __init__(self, path):
        """Read file and read first file and trans-unit nodes."""
        
        self.docXLIFF = parse(path)
        self.xliffRoot = self.docXLIFF.documentElement
        cursor = self.xliffRoot
        self.fileNode = self._first_child(cursor, 'file')
        self.bodyNode = self._first_child(self.fileNode, 'body') 
        self.transunit = self._first_child(self.bodyNode, 'trans-unit')
        
    def _first_child(self, cursor, name):
        """Return cursor's first child element called name."""
        
        cursor = cursor.firstChild
        while True:
            if cursor.nodeName == name:
                break
            else:
                cursor = cursor.nextSibling
        return cursor

    def _last_child(self, cursor, name):
        """Return cursor's last child element called name."""

        cursor = cursor.lastChild
        while True:
            if cursor.nodeName == name:
                break
            else:
                cursor = cursor.previousSibling
        return cursor

    def _next_sibling(self, cursor, name):
        """Return cursor's next sibling called name."""
        
        cursor = cursor.nextSibling
        while True:
            if cursor.nodeName == name:
                break
            else:
                cursor = cursor.nextSibling
        return cursor
    
    def _prev_sibling(self, cursor, name):
        """Return cursor's previous sibling called name."""
        
        cursor = cursor.previousSibling
        while True:
            if cursor.nodeName == name:
                break
            else:
                cursor = cursor.previousSibling
        return cursor
    
    def next_transunit(self):
        """Iterate to next trans-unit. Go to next file node if necessary.
        
        If the current node is the last trans-unit in the document,
        next_transunit() stays in the node.
        """

        last = self._last_child(self.bodyNode, 'trans-unit')
        if self.transunit == last:
            # This is the last child of the current file
            self.next_file()
        else:
            self.transunit = self._next_sibling(self.transunit, 'trans-unit')
            
    def prev_transunit(self):
        """Iterate to previous trans-unit. Go to previous file node if
        necessary.
        
        If the current node is the first trans-unit in the document,
        prev_transunit() stays in the node.
        """

        first = self._first_child(self.bodyNode, 'trans-unit')
        if self.transunit == first:
            self.prev_file()
        else:
            self.transunit = self._prev_sibling(self.transunit, 'trans-unit')
    
    def next_file(self):
        """Iterate to next file node.
        
        Set trans-unit to the first trans-unit in the current file. If the
        current node is the last file node, next_file does nothing.
        """
        
        last = self._last_child(self.xliffRoot, 'file')
        if self.fileNode == last:
            # We are already in the last file
            return
        else:
            # Update cursor, including file, body and first trans-unit
            self.fileNode = self._next_sibling(self.fileNode, 'file')
            self.bodyNode = self._first_child(self.fileNode, 'body')
            self.transunit = self._first_child(self.bodyNode, 'trans-unit')
            
    def prev_file(self):
        """Iterate to previous file node.
        
        Set trans-unit to the last trans-unit in the current file. If the
        current node is the first file node, next_file does nothing.
        """
        first = self._first_child(self.xliffRoot, 'file')
        if self.fileNode == first:
            return
        else:
            # Start from the bottom
            self.fileNode = self._prev_sibling(self.fileNode, 'file')
            self.bodyNode = self._last_child(self.fileNode, 'body')
            self.transunit = self._last_child(self.bodyNode, 'trans-unit')

    def read_languages(self):
        """Return language pair."""

        slang = self.file.getAttribute("source-language")
        tlang = self.file.getAttribute("target-language")
        return (slang, tlang)
  
    def read_source(self, node=None):
        """Return source text of current trans-unit."""
        ret = ""
        if not node:
            node = self.transunit
        try:
            sourceNode = self._first_child(node, "source")
        except AttributeError:
            return ret
        childNodes = sourceNode.childNodes
        for n in childNodes:
            ret += n.toxml()
        ret.strip()
        return ret
    
    def read_target(self, node=None):
        """Return source text of current trans-unit."""
        ret = ""
        if not node:
            node = self.transunit
        try:
            targetNode = self._first_child(node, "target")
        except AttributeError:
            return ret
        childNodes = targetNode.childNodes
        for n in childNodes:
            ret += n.toxml()
        ret.strip()
        return ret
    
    def read_altTrans(self):
        """Return a list of dictionaries with source, target, and quality keys."""

        ret = []
        altTransNodes = self.transunit.getElementsByTagName("alt-trans")
        for alt in altTransNodes:
            dic = {}
            target = self.read_target(alt)
            if not target:
                continue
            dic['target'] = target
            if alt.hasAttribute("match-quality"):
                dic['quality'] = alt.getAttribute("match-quality")
            source = self.read_source(alt)
            if source:
                dic['source'] = source
            ret.append(dic)
        return ret    
  
    def edit_target(self, ttext, state=None, altTrans=None):
        """Edit target node of current trans-unit and add alt-trans nodes."""
        
        newTarget = self.write_target(ttext, state)
        try:
            prevTarget = self._first_child(self.transunit, "target")
        except AttributeError:
            self.transunit.appendChild(newTarget)
        else:
            self.transunit.replaceChild(newTarget, prevTarget)
        # [NOTE]
        # I believe it is safer (or just easier) to delete all the previous 
        # alt-trans and create new nodes.
        if not altTrans:
            return
        prevAlt = self.transunit.getElementsByTagName("alt-trans")
        for alt in prevAlt:
            retNode = self.transunit.removeChild(alt)
            retNode.unlink()
        newAltTrans = self.write_altTrans(altTrans)
        for n in newAltTrans:
            self.transunit.appendChild(n)

# [TEST]
TESTXML = """<?xml version="1.0"?>
   <xliff version="1.1">
    <file source-language="en" datatype="plaintext" 
          original="test" target-language="es">
     <body>
      <trans-unit id="1">
       <source>First they ignore you.</source>
       <target state="final">Primero te ignoran</target>
      </trans-unit>
      <trans-unit id="2">
       <source>Then they laugh at you.</source>
       <alt-trans match-quality="80%">
        <source>they laugh at me</source>
        <target>Se rieron de me</target>
       </alt-trans> 
      </trans-unit>
      <trans-unit id="3">
       <source>Then they fight you</source>
      </trans-unit>
      <trans-unit id="4">
       <source>Then you win.</source>
      </trans-unit>
      <trans-unit id="5">
       <source>Mohandas Gandhi</source>
      </trans-unit>
     </body>
    </file>
   </xliff>
"""
import unittest
import os, os.path
class ConfigTest(unittest.TestCase):
    def setUp(self):
        self.path = 'test.miau'
        self.slang = 'en'
        self.tlang = 'es'
        self.text1 = "First <b>they</b> ignore you."
        self.text2 = "Then they laugh at you."
        self.text3 = "Then they fight you."
        self.text4 = "Then you win."
        self.text5 = "Mohandas Gandhi"
                
    def tearDown(self):
        if os.path.isfile(self.path):
            os.unlink(self.path)
    
    def test_01(self):
        """ 1. Testing WriteXLIFF class"""
        obj = WriteXLIFF()
        obj.write_file("plaintext", "testFile", self.slang, self.tlang)
        obj.write_transunit("1", self.text1)
        obj.write_transunit("2", self.text2)
        obj.write_transunit("3", self.text3)
        obj.write_transunit("4", self.text4)
        obj.write_transunit("5", self.text5)
        fh = open(self.path, "w")
        fh.write(obj.docXLIFF.toxml())
        fh.close()
        length = len(obj.docXLIFF.toxml())
        self.failUnless(length > 0, 'File was not created')
        
    def test_02(self):
        """Test ReadXLIFF"""
        fh = open(self.path, "w")
        fh.write(TESTXML)
        fh.close()
        a = ReadXLIFF(self.path)
        self.failUnless(a.fileNode.nodeName == 'file', "Incorrect file name")
        a.next_transunit()
        self.failUnless(a.transunit.getAttribute("id") == "2", "Incorrect id")
        a.prev_transunit()
        self.failUnless(a.transunit.getAttribute("id") == "1", "Incorrect id")
        while True:
            b = a.transunit
            a.next_transunit()
            s = a.read_source()
            self.failUnless(len(s) > 0, "Source was not read properly")
            if a.transunit == b:
                break

if __name__ == '__main__':
    unittest.main()
