### Copyright (C) 2006-2007 Manuel Ospina <mospina@redhat.com>

# This file is part of miaow.
#
# miaow is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# miaow is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with miaow; if not, write to the Free Software
# Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA

# -*- encoding: utf-8 -*-

"""XML to l10n data

Extract and merge translatable data from xml files.
Based on xml2po by Danilo Segan <danilo@kvota.net>.
"""

import sys
import libxml2
import os, os.path
import l10nData

class Extract:
    
    """
    Extract translatable text from a XML document.
    
    Usage:
    
    x = Extract(finalTags=None, ignoreTags=None)
    l20nObj = x.extract(filename)
    
    finalTags - list of tags that should be considered as final tags.
    ignoreTags - list og tags that should be ignored.
    filename - path to xml file
    
    The method extractor returns a L10nData object that can be save latter on as
    a XLIFF or PO file.
    """
    
    def __init__(self, finalTags=None, ignoreTags=None):
        """Initialize final and ignore tag list."""
    
        self.filename = ''
        self.tuid = 0
        self.l10n = l10nData.L10nData()
        if finalTags and not isinstance(finalTags, list):
            raise TypeError, "finalTags must be a list or None"
            sys.exit(1)
        self.finalTags = finalTags
        if ignoreTags and not isinstance(ignoreTags, list):
            raise TypeError, "ignoreTags must be a list or None"
            sys.exit(1)
        self.ignoreTags = ignoreTags
       
    def extract(self, filename):
        """Return a L10nData object with the translatable text of filename."""
        
        if not os.path.isfile(filename):
            raise IOError, "file not found"
            sys.exit(1)
        self.filename = filename

        f = l10nData.L10nFile()
        f.datatype = "plaintext"
        f.original = os.path.abspath(self.filename)
        f.slang = "en-US"
        self.l10n.add_file(f)

        ctxt = libxml2.createFileParserCtxt(self.filename)
        ctxt.replaceEntities(1)
        ctxt.lineNumbers(1)
        ctxt.parseDocument()
        doc = ctxt.doc()
        assert doc.name == self.filename

        self.doSerialize(doc)
        
        return self.l10n

    def doSerialize(self, node):
        """Serialize node and its children depending of its type."""
        
        if self._ignoreNode(node):
            return ''
        elif not node.children:
            return node.serialize("utf-8")
        elif node.type == 'entity_ref':
            # [NOTE]
            # xml2po check whether to keep or replace entities. Here, entities
            # are not replaced.
            return node.serialize('utf-8')
        elif node.type == 'entity_decl':
            return node.serialize('utf-8')
        elif node.type == 'text':
            return node.serialize('utf-8')
        elif node.type == 'element':
            return self._processElementTag(node)
        else:
            child = node.children
            outtxt = ''
            while child:
                outtxt += self.doSerialize(child)
                child = child.next
            return outtxt

    def _ignoreNode(self, node):
        """Return whether node is in ignoreTags list or not."""
        
        if not self.ignoreTags:
            if node.type in ('dtd', 'comment'):
                return True
            else:
                return False
        else:
            if node.name in self.ignoreTags or node.type in ('dtd', 'comment'):
                return True
            return False

    def _isFinalNode(self, node):
        """Return whether node is consider final."""
         
        if not self.finalTags:
            auto = self._autoNodeIsFinal(node)
            # Check if any of the parents is also autoNodeIsFinal,
            # and if it is, don't consider this node a final one.
            parent = node.parent
            while parent:
                if self._autoNodeIsFinal(parent):
                    auto = False
                    break
                parent = parent.parent
            return auto
        if node.type == 'element' and node.name in self.finalTags:
            return True
        return False

    def _autoNodeIsFinal(self, node):
        """Return whether a node should be considered final.
        
        A final node is a text node.
        """

        final = False 
        if node.isText() and node.content.strip() != '':
            return True
        child = node.children
        while child:
            if child.type in ['text'] and  child.content.strip() != '':
                final = True
                break
            child = child.next
        return final
        
    def _processElementTag(self, node):
        """Process node of type element."""
        
        assert node.type == 'element'
        final = self._isFinalNode(node)
        outtxt = ''
        child = node.children
        while child:
            result = self.doSerialize(child)
            outtxt += result
            child = child.next
        if final:
            outtxt = self._processFinalTag(node, outtxt)
        else:
            if self._worthOutputting(node):
                outtxt = self._processFinalTag(node, outtxt)
        # [DEBUG]
        #return '<%s>%s</%s>' % (self._startTagForNode(node), outtxt, node.name)
        # [NOTE]
        # Why does it not return node.serialize("utf8") ?
        return '<%s>%s</%s>' % (node.name, outtxt, node.name)
        # -------

    def _processFinalTag(self, node, outtxt):
        """Create the translation unit for node."""

        if not self._isFinalNode(node):
            return outtxt
        if not self._ignoreNode(node):
            self.tuid += 1
            tu = l10nData.TU()
            tu.source = outtxt
            tu.id = self.tuid
            self.l10n.add_tu(tu)
            # [TODO]
            # We need to implement a <note> node in xliff.py to include the
            # returning value of:
            #   * node.lineNo(), 
            #   * self._getCommentForNode(node) 
            #   * self._isSpacePreserveNode(node)
        return outtxt
    
    def _worthOutputting(self, node):
        """Returns True if node is "worth outputting", otherwise False.
        
        Node is "worth outputting", if none of the parents isFinalNode, and 
        it contains non-blank text and entities.
        """
        worth = True
        parent = node.parent
        while parent:
            if self._isFinalNode(parent):
                worth = False
                break
            parent = parent.parent
        if not worth:
            return False
        return self._autoNodeIsFinal(node)

        # [METHODS]
        # self._getCommentForNode(node) ?
        # self._isSpacePreserveNode(node)
        # self._startTagForNode(node) ?

class Merge(Extract):    
    """
    Merge translated text into a XML document.

    Usage:

    x = Merge(lData, finalTags=None, ignoreTags=None)
    xmlDoc = x.merge(filename)

    lData - the l10n data object that contains the translations
    finalTags - list of tags that should be considered as final tags.
    ignoreTags - list og tags that should be ignored.
    filename - path to xml file

    The method merger returns a string with the xmlDoc.
    """
    
    def __init__(self, lData, finalTags=None, ignoreTags=None):
        """Initialize final and ignore tag list."""

        self.filename = ''
        if not isinstance(lData, l10nData.L10nData):
            raise TypeError, "l10nData should be a L10nData instance"
            sys.exit(1)
        self.l10n = lData
        if finalTags and not isinstance(finalTags, list):
            raise TypeError, "finalTags must be a list or None"
            sys.exit(1)
        self.finalTags = finalTags
        if ignoreTags and not isinstance(ignoreTags, list):
            raise TypeError, "ignoreTags must be a list or None"
            sys.exit(1)
        self.ignoreTags = ignoreTags
       
    def merge(self, filename):
        """Return the translated xml document."""
        
        if not os.path.isfile(filename):
            raise IOError, "file not found"
            sys.exit(1)
        self.filename = filename
        ctxt = libxml2.createFileParserCtxt(self.filename)
        ctxt.lineNumbers(1)
        ctxt.parseDocument()
        doc = ctxt.doc()
        assert doc.name == self.filename

        self.doSerialize(doc)

        return doc.serialize("utf8")

    def _processFinalTag(self, node, outtxt):
        """Create the translation unit for node."""
        
        if not self._isFinalNode(node):
            return outtxt
        if not self._ignoreNode(node):
            # [TODO]
            # Handle space preserve nodes and lineNo
            tu = self.l10n.search_source(outtxt)
            # [BUG]
            # Why is not reading tu properly?
            if not tu:
                return outtxt
            #-------------------------------
            newtxt = tu.target
            if not newtxt:
                newtxt = outtxt
            newNodeTxt = node.serialize("utf8").replace(outtxt, newtxt)
            newNodeDoc = libxml2.parseMemory(newNodeTxt, \
                                             len(newNodeTxt))
            newNode = newNodeDoc.getRootElement()                                             
            if node:
                next = node.next
                node.replaceNode(newNode.copyNodeList())
                node.next = next
        return newtxt

# [TEST]
XMLTEXT ="""<?xml version="1.0"?>
 <miaow-test>
  <title>Miaow Test</title>
   <p>First they <b>ignore</b> you.</p>
   <p><t>Then they laugh at you.</t></p>
   <p id="10">Then they <i>fight</i> you</p>
   <p>Then you win.<end /></p>
   <author>Mohandas Gandhi</author>
  </miaow-test>
"""
POFILE = """
msgid ""
msgstr ""

msgid "Miaow Test"
msgstr "Test de Miaow"

msgid "First they <b>ignore</b> you."
msgstr "Primero te <b>ignoran</b>."

msgid "Then they laugh at you."
msgstr "Luego se rién de tí"

msgid "Then they <i>fight</i> you"
msgstr ""

msgid "Then you win.<end/>"
msgstr "Luego vences."

msgid "Mohandas Gandhi"
msgstr ""
"""
import unittest
class ConfigTest(unittest.TestCase):
    def setUp(self):
        self.xml = "test-l10ndata.xml"
        self.po = "text-l10ndata.po"

    def tearDown(self):
        if os.path.isfile(self.xml):
            os.unlink(self.xml)
        if os.path.isfile(self.po):
            os.unlink(self.po)

    def test_01(self):
        """1. Testing Extract without final and ignore tag list."""
        fh = open(self.xml, "w")
        fh.write(XMLTEXT)
        fh.close()
        m = Extract()
        l = m.extract(self.xml)
        self.failUnless(len(l.l10nFiles[0].tu) > 0, 'Translation units no created.')
        #l.write(self.po)

    def test_02(self):
        """2. Testing Extract with final tag list."""
        fh = open(self.xml, "w")
        fh.write(XMLTEXT)
        fh.close()
        m = Extract(['p'])
        self.failUnless(len(m.finalTags) == 1, 'Wrong number of finalTags.')
        l = m.extract(self.xml)
        self.failUnless(len(l.l10nFiles[0].tu) == 4, 'TU = %d, 4 expected.' \
                                                     % len(l.l10nFiles[0].tu))

    def test_03(self):
        """3. Testing Extract with ignore tag list."""
        #self.failUnless(os.path.isfile(self.path), 'File was not created.')
        fh = open(self.xml, "w")
        fh.write(XMLTEXT)
        fh.close()
        m = Extract(ignoreTags = ['author'])
        self.failUnless(len(m.ignoreTags) == 1, 'Wrong number of ignoreTags.')
        l = m.extract(self.xml)
        self.failUnless(len(l.l10nFiles[0].tu) == 5, \
                        'TU = %d, 5 expected.' % len(l.l10nFiles[0].tu))
        while True:
            self.failUnless(len(l.tu.source) > 0, "Source is empty")
            pTu = l.cTu
            l.prev()
            if pTu == l.cTu:
                break

    def test_04(self):
        """4. Testing Merge without final and ignore tag list."""
        fh = open(self.xml, "w")
        fh.write(XMLTEXT)
        fh.close()
        fh = open(self.po, "w")
        fh.write(POFILE)
        fh.close()
        l = l10nData.L10nData(self.po, "po")
        m = Merge(l)
        t = m.merge(self.xml)
        self.failUnless(len(t) > 0, 'xml text was not returned.')

if __name__ == '__main__':
    unittest.main()
