### Copyright (C) 2006-2007 Manuel Ospina <mospina@redhat.com>

# This file is part of miaow.
#
# miaow is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# miaow is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with miaow; if not, write to the Free Software
# Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA

"""Project management tool

This module defines functions and classes to create and manipulate localization
projects.
"""
# [NOTE]
# Terms:
# slang = source language
# tlang = target language
# template = file containing all the translatable string of a document.

# [TODO]
# 1. Error handler for ProjectReader
#	try:
#		...
#	except ConfigParser.NoOptionError:
import sys
import os, os.path
import ConfigParser
import glob

class ProjectReader:
	
	"""
	Read a project file:
	
	Methods:
	
	__init__(projectPath)
	#Read_project()
	get_source_lang()
	get_languages()
	get_initializer()
	get_file_formats()
	get_files(fileFormat)
	get_extractor(fileFormat)
	get_merger(fileFormat)

	The format of a project file is:
	
	------------------------------------------
	[project]
	url = projectDir
	source_language = lang
	target_languages = lang, lang
	name = projectName
	
	[fileFormat]
	files = file, file 
	extractor = app
	merger = app
	------------------------------------------
	"""

	def __init__(self, projectPath):
		"""Set global attributes and read project file
		
		*projectPath* is the full path to the project file (dir/file).
		This method defines projectDir and projectName from the projectPath. 
		"""
		self.projectPath = projectPath
		self.projectDir = os.path.split(projectPath)[0]
		self.projectName = os.path.split(projectPath)[1]
		self.config = ConfigParser.ConfigParser()
		
		self.config.read(self.projectPath)
		
	def get_source_lang(self):
		"""Return source language"""
		ret = None
		try:
			ret = self.config.get("project", "source_language")
		except ConfigParser.NoOptionError:
			print "Source language has not been set"
			sys.exit(1)
		return ret
		
	def get_languages(self):
		"""Return list of target languages"""
		try:
			targets = self.config.get("project", "target_languages")
		except ConfigParser.NoOptionError:
			targets = ''
		ret = [ x.strip() for x in targets.split(",")]
		ret = filter(lambda x: len(x) > 0, ret)
		return ret
	
	def get_initializer(self):
		"""Return the initializer app."""
		ret = None
		try:
			ret = self.config.get("project", "initializer")
		except ConfigParser.NoOptionError:
			print "Initializer has not been set"
			sys.exit(1)
		return ret
		
	def get_file_formats(self):
		"""Return files formats in the current project"""
		fileFormat = self.config.sections()
		# First section is "project"... get rid of it
		fileFormat.pop(0)
		return fileFormat
		
	def get_files(self, fileFormat):
		"""Return files in a specified file format"""
		ret = []
		fileList = self.config.get(fileFormat, "files")
		pre = [ x.strip() for x in fileList.split(",")]
		pre = filter(lambda x: len(x) > 0, pre)
		for filePath in pre:
			f = glob.glob(filePath)
			ret.extend(f)
		return ret
		
	def get_extractor(self, fileFormat):
		"""Return extractor for a specified file format"""
		ret = None
		try:
			ret = self.config.get(fileFormat, "extractor")
		except ConfigParser.NoOptionError:
			print "Extractor has not been set"
			sys.exit(1)
		return ret
		
	def get_merger(self, fileFormat):
		"""Return merger for a given file Format"""
		ret = []
		try:
			ret = self.config.get(fileFormat, "merger")
		except ConfigParser.NoOptionError:
			print "Merger has not been set"
			sys.exit(1)
		return ret
		
class ProjectWriter:
	
	"""
	Write a project file:
	
	Methods:
	
	__init__(projectPath)
	create_project()
	add_source_lang(lang)
	add_language(lang)
	add_initializer()
	add_file(fileFormat, file)
	add_extractor(fileFormat, app)
	add_merger(fileFormat, app)
	save()

	The format of a project file is:
	
	------------------------------------------
	[project]
	url = projectDir
	source_language = lang
	target_languages = lang, lang
	name = projectName
	
	[fileFormat]
	files = file, file 
	extractor = app
	merger = app
	------------------------------------------
	"""

	def __init__(self, projectPath):
		"""Set global attributes and read project file
		
		*projectPath* is the full path to the project file (dir/file).
		This method defines projectDir and projectName from the projectPath. 
		"""
		self.projectPath = projectPath
		self.projectDir = os.path.split(projectPath)[0]
		self.projectName = os.path.split(projectPath)[1]
		self.config = ConfigParser.ConfigParser()
		# If the file exists, read it.
		self.config.read(self.projectPath)
		
	def create_project_template(self):
		"""Create a template of a pmt file."""
		
		if not self.config.has_section('project'):
			self.config.add_section('project')
		self.config.set('project', 'source_language', '')
		self.config.set('project', 'target_languages', '')
		self.config.set('project', 'initializer', '')
		
		self.config.add_section('fileFormat')
		self.config.set('fileFormat', 'files', '')
		self.config.set('fileFormat', 'extractor', '')
		self.config.set('fileFormat', 'merger', '')
		
	def add_source_lang(self, lang):
		"""Add the source language to the project configuration
		
		The section [project] should exists from a previous call to
		create_project(); if the section doesn't exist, it is created. 
		"""
		if not self.config.has_section("project"):
			self.config.add_section("project")
		self.config.set("project", "source_language", lang)
		
	def add_language(self, lang):
		"""Add a language to the project configuration.
		
		Loop through the language list to check whether the language exists.
		"""
		if not self.config.has_section("project"):
			self.config.add_section("project")	
		# If the option doesn't exist, create it and exit
		if not self.config.has_option("project", "target_languages"):
			self.config.set("project", "target_languages", lang)
			return
		# else:	
		targetLanguages = self.config.get("project", "target_languages")
		# Get list of languages
		langList = [ x.strip() for x in targetLanguages.split(",")]
		# Get rid off empty items
		langList = filter(lambda x: len(x) > 0, langList)
		for eachLang in langList:
			if eachLang == lang:
				return
		langList.append(lang)
		targetLanguages = ",".join(langList)
		self.config.set("project", "target_languages", targetLanguages)
	
	def add_initializer(self, app):
		"""Add the inicializer app.

		A initializer is an app that create the working files for each language
		from a given template. The initializer should received the option --lang
		(-l) and a template file as a argument. The initialized file is output
		to stdout. e.x.:

		initializer --lang es_ES template > file_es

		Initializer can be called using the initialize() function
		"""
		self.config.set("project", "initializer", app)
		
	def add_file(self, fileFormat, file):
		"""Add a file to the project file.
		
		Create a fileFormat section if necessary. Check if the file is already
		in the list.
		"""		
		# If the section doesn't exist, create it, add file and exit.
		if not self.config.has_section(fileFormat):
			self.config.add_section(fileFormat)
			self.config.set(fileFormat, "files", file)
			return
		# else:	
		files = self.config.get(fileFormat, "files")
		# Get list of files
		fileList = [ x.strip() for x in files.split(",")]
		# Get rid off empty items
		fileList = filter(lambda x: len(x) > 0, fileList)		
		for eachFile in fileList:
			if eachFile == file:
				return
		fileList.append(file)
		files = ",".join(fileList)
		self.config.set(fileFormat, "files", files)
	
	def add_extractor(self, fileFormat, app):
		"""Add the extractor app to use with the files in the fileFormat
		section.

		A extractor is a filter that extract the translatable sections of a
		document. The extractor should receive a file as an argument and send
		the output to stdout. e.x:

		extractor file.xml > file.xliff

		The method extract(extractor, source) is used to call this app.
		"""
		
		# Check if section exist.
		if not self.config.has_section(fileFormat):
			self.config.add_section(fileFormat)		
		self.config.set(fileFormat, "extractor", app)
		
	def	add_merger(self, fileFormat, app):
		"""Add the merger app to use with the file in the fileFormat
		section.

		A merger is a filter that generate the translated document from a
		translation file. The merger should receive the original file as a first
		argument and the translation file (xliff/po) as an argument; the
		output is sent to stdout. e.x.:
		
		mergerScript file.xml translation.xliff > translatedFile.xml 

		The method merge(merger, source, trans) is used to call this app.
		"""
		
		if not self.config.has_section(fileFormat):
			self.config.add_section(fileFormat)
		self.config.set(fileFormat, "merger", app)
	

	def save(self):
		"""Save project configuration to disk."""
		fh = file(self.projectPath, "w")
		self.config.write(fh)
		fh.close()
	

class PMT(ProjectReader, ProjectWriter):
	
	"""
	API for manipulating project files.
	
	Methods:
	
	update()
	update_templates()
	update_language(lang)
	
	Usage:
	 p = PMT(path)
	 p.create_project()
	 p.add_source_lang(lang)
	 p.get_source_lang()
	 p.save()

	 path is the full path to the project file. If the file doesn't exist can e
	 created using create_project() and then save to disk with save().
	"""

	def __init__(self, projectPath):
		ProjectReader.__init__(self, projectPath)
		ProjectWriter.__init__(self, projectPath)
	
	def update(self, lang=None):
		"""Update project infrastructure."""
		
		# Project file
		if not os.path.isfile(self.projectPath):
			self.save()
		# Create or update templates
		self.update_templates()
		# Target language directories
		if not lang:
			tlang = self.get_languages()
		else:
			tlang = [lang]
		for eachLang in tlang:
			self.update_language(eachLang)
	
	def update_templates(self):
		"""Update the template files of the project"""
		
		slang = self.get_source_lang()
		slangDir = os.path.join(self.projectDir, slang)
		if not os.path.isdir(slangDir):
			os.makedirs(slangDir)
		formats = self.get_file_formats()
		for eachFormat in formats:
			extractor = self.get_extractor(eachFormat)
			files = self.get_files(eachFormat)
			for eachFile in files:
				# [Note]
				# Only POT format is supported in this prototype. The aimed
				# formar is XLIFF. 
				fileName = os.path.basename(eachFile).split('.')[0] + '.pot'
				fileName = os.path.join(slangDir, fileName)
				fd = open(fileName, "w")
				fd.write(extract(extractor, eachFile))
				fd.close()
				
	def update_language(self, lang):
		"""Update translation files for the specified language.
		
		This method does not update the templates.
		"""

		# [Note]
		# The initialization is done comparing the content of a
		# language directory against the list of files in slang. 
		slang = self.get_source_lang()
		slangDir = os.path.join(self.projectDir, slang)
		langDir = os.path.join(self.projectDir, lang)
		if not os.path.isdir(langDir):
			os.makedirs(langDir)
		# Initialize files each language
		initializer = self.get_initializer()
		for fileName in os.listdir(slangDir):
			basename = fileName.split(".")[0]
			template = os.path.join(slangDir, fileName)
			transfile = os.path.join(langDir, basename + '.po') 
			fd = open(transfile, 'w')
			fd.write(initialize(template, lang, initializer))
			fd.close()	
			
	def build(self, lang=None, path=None):
		"""Build target files."""
		
		# Target language directories
		if not lang:
			tlang = self.get_languages()
		else:
			tlang = [lang]
		for eachLang in tlang:	
			self.build_language(eachLang, path)
				
	def build_language(self, lang, path=None):
		"""Build target files for the specified language."""
		if not path:
			path = 'target'
		path = os.path.join(self.projectDir, path + '-' + lang)
		if not os.path.isdir(path):
			os.makedirs(path)
		langDir = os.path.join(self.projectDir, lang)
		# Build files language
		formats = self.get_file_formats()
		for eachFormat in formats:
			merger = self.get_merger(eachFormat)
			files = self.get_files(eachFormat)
			for eachFile in files:
				tmpName = os.path.basename(eachFile).split('.')
				basename = tmpName[0]
				extention = tmpName[-1]
				transfile = os.path.join(langDir, basename + '.po')
				outFile = '%s-%s.%s' % (basename, lang,	extention)
				output = os.path.join(path, outFile)
				fd = open(output, 'w')
				fd.write(merge(merger, eachFile, transfile))
				fd.close()
				
# [NOTE] 
# In this prototype the initialize, the extractor and the merger
# should be given. In the future, such options must be optionals; the
# defaults values will be read from an .rc file.
def extract(extractor, source):
	"""Call the extractor app."""
		
	ret = ''
	filter = '%s %s' % (extractor, source)
	miPipe = os.popen(filter)
	ret = miPipe.read()
	miPipe.close()
	return ret
		
def merge(merger, source, trans):
	"""Call the merger app."""		
	ret = ''
	filter = '%s %s %s' % (merger, source, trans)
	miPipe = os.popen(filter)
	ret = miPipe.read()
	miPipe.close()
	return ret

def initialize(template, tlang, initializer):
	"""Initialize a file in the target language."""
	# [NOTE]
	# A initializer can:
	#	1. initialize a file
	#	2. pretranslate the initialized file.
	ret = ''
	# [NOTE]
	# msginit is not a good initializer according to PMT. The input must be
	# given by the option -i. For PMT, the initializer should be called as:
	#
	#filter = '%s --lang %s %s' % (initializer, tlang, template)
	filter = '%s -l %s -i %s' % (initializer, tlang, template)
	#filter = 'msginit --no-translator -l %s -i %s -o -' % (tlang, template)
	miPipe = os.popen(filter)
	ret = miPipe.read()
	miPipe.close()
	return ret

# [TEST]
# ======
import unittest
import os, os.path
class ConfigTest(unittest.TestCase):
	def setUp(self):
		self.sourceLangs = "en"
		self.targetLangs = ["es", "ru"]
		self.project = "/junk/mospina/BlueCAT/prototype/miau/miau/tmp/test-1.cat"
		self.myFile = "/junk/mospina/BlueCAT/prototype/miau/miau/tmp/tutorial.xml"
						
	def tearDown(self):
		# Need to read files before attempting to delete dirs
		if os.path.isfile(self.project):
			os.unlink(self.project)
		pass
	def test_01(self):
		"""1. Testing Writer"""
		p = PMT(self.project)
		p.add_source_lang("en") 
		for lang in self.targetLangs:
			p.add_language(lang)
		p.add_initializer("msginit --no-translator -o -")	
		p.add_file("DocBook", self.myFile)
		p.add_extractor("DocBook", "xml2po")
		p.add_merger("DocBook", "xml2po -r")
		p.save()
		self.failUnless(os.path.isfile(self.project), 'File was not created')
	
	def test_02(self):
		"""1. Testing Reader"""
		
		self.test_01()
		p = PMT(self.project)
		print p.get_source_lang()
		t = p.get_languages()
		print str(t)
		s = p.get_file_formats()
		print s
		for each in s:
			print "Ext", p.get_extractor(each)
			print "Merger", p.get_merger(each)
	
	def test_03(self):
		"""3. Testing Update"""
		
		self.test_01()
		p = PMT(self.project)
		p.update()
		slang = p.get_source_lang()
		slang = os.path.join(p.projectDir, slang)
		self.failUnless(os.path.isdir(slang), 'Source Lang. dir was not created')
		tl = p.get_languages()
		s = p.get_file_formats()
		file = p.get_files(s[0])[0]
		file = os.path.basename(file).split('.')[0] + '.po'
		for t in tl:
			tlangFile = os.path.join(t, file)
			tlangFile = os.path.join(p.projectDir, tlangFile)
			self.failUnless(os.path.isfile(tlangFile), 'TFile was not created')
	
	def test_04(self):
		"""4. Testing Build"""
		
		self.test_01()
		p = PMT(self.project)
		p.build()
		for lang in p.get_languages():
			target = os.path.join(p.projectDir, 'target-' + lang)
			self.failUnless(os.path.isdir(target), 'Target dir was not created')

if __name__ == '__main__':
	unittest.main()
