### Copyright (C) 2006-2007 Manuel Ospina <mospina@redhat.com>

# This file is part of miaow.
#
# miaow is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# miaow is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with miaow; if not, write to the Free Software
# Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA

# -*- coding: utf8 -*-
"""
tm is a Translation Memory DataBase handler
"""

from sqlite import connect

class TMReader:
    
    """
    Read data from the Translation Memory Database.

    Methods:

    get_languages()
    get_slang()
    get_tlangs()
    search_entry(tlang, msg)
    match_entry(tlang, msg)
    """

    def __init__(self, connection, cursor):
        self.connection = connection
        self.cursor = cursor

    def get_languages(self):
        """Return list of languages."""
        
        query = """
        SELECT langcode, langname, adminlang 
        FROM languages ORDER BY langname
        """
        self.cursor.execute(query)
        result = self.cursor.fetchall()
        return result 
    
    def get_slang(self):
        """Return the source language code."""
        
        query = """
        SELECT langcode FROM languages WHERE adminlang = 'True'
        """
        self.cursor.execute(query)
        tmp = self.cursor.fetchone()
        if tmp:
            return tmp[0]
        else:
            raise SlangDoesNotExist

    def get_tlangs(self):
        """Return list of target language codes."""

        ret = []
        query = """
        SELECT langcode FROM languages WHERE adminlang = 'False'
        """
        self.cursor.execute(query)
        tmp = self.cursor.fetchall()
        if not tmp:
            return None
        for item in tmp:
            ret.append(item[0])
        return ret

    def _get_langcode(self, langname):
        """Return the code of a given language. 

        'langname' is the name of the language. If the code of the language is
        given instead, the method returns the code anyway.
        """

        query = """
        SELECT langcode FROM languages 
        WHERE langname LIKE '%s' OR langcode LIKE '%s'
        """ % (langname, langname)
        self.cursor.execute(query)
        tmp = self.cursor.fetchone()
        if tmp:
            return tmp[0]
        else:
            raise LookupError
    
    def _get_langname(self, langcode):
        """Return the name of a given language. 

        'langcode' is the name of the language. If the name of the language is
        given instead, the method returns the code anyway.
        """

        query = """
        SELECT langname FROM languages 
        WHERE langcode LIKE '%s' OR langname LIKE '%s'
        """ % (langcode, langcode)
        self.cursor.execute(query)
        tmp = self.cursor.fetchone()
        if tmp:
            return tmp[0]
        else:
            raise LookupError

    def search_entry(self, tlang, msg):
        """Lookup for a exact match. Return target if found."""
        
        # Preconditions:
        tlang = self._get_langcode(tlang)
        msg = escape_msg(msg)
        msgid = self._get_msgid(msg)
        if not msgid:
            return None
        # else:     
        tmsg = self._get_target(msgid, tlang)
        return tmsg
        
    def match_entry(self, tlang, msg):
        """Look up for similar matches. Return list of targets."""
        
        # Preconditions:
        tlang = self._get_langcode(tlang)
        
        msg = escape_msg(msg)
        query = """
        SELECT msgid FROM source
        WHERE msg LIKE '%s'
        """ % msg
        self.cursor.execute(query)
        tmp = self.cursor.fetchall()
        if not tmp:
            return None
        # else: 
        ret = []
        for msgid in tmp:
            tmsg = self._get_target(msgid[0], tlang)
            if not tmsg:
                continue
            ret.append(tmsg)    
        return ret
    
    def _get_msgid(self, msg):
        """Return source's ID."""
        
        query = """
        SELECT msgid FROM source
        WHERE msg = '%s'
        """ % msg
        self.cursor.execute(query)
        tmp = self.cursor.fetchone()
        if tmp:
            return tmp[0]
        else:   
            return None
        
    def _get_target(self, msgid, tlang):
        """Return target for the given ID and target language."""
        
        # Preconditions:
        tlang = self._get_langcode(tlang)
        msgid = int(msgid)
        
        query = """
        SELECT tmsg FROM target
        WHERE msgid = %d AND langcode = '%s'
        """ % (msgid, tlang)
        self.cursor.execute(query)
        tmp = self.cursor.fetchone()
        if tmp:
            return tmp[0]
        else:
            return None

class TMWriter(TMReader):
    
    """
    Write to the Translation Memory Database.
        
    Methods:

    add_slang(name, code)
    add_tlang(name, code)
    add_entry(tlang, source, target)
    """

    def __init__(self, connection, cursor):
        self.connection = connection
        self.cursor = cursor
        TMReader.__init__(self, self.connection, self.cursor)

    def add_slang(self, name, code):
        """Add the source language to the DataBase.
        
        The source language must be set only once.
        """
        
        try:
            self.get_slang()
            return
        except SlangDoesNotExist:
            pass
        query = """
        INSERT INTO languages(langcode, langname, adminlang)
        VALUES ('%s', '%s', 'True')
        """ % (code, name)
        self.cursor.execute(query)
        self.connection.commit()

    def add_tlang(self, name, code):
        """Add a target language to the DataBase."""
        
        try:
            self._get_langcode(code)
            return
        except LookupError:
            pass
        query = """
        INSERT INTO languages(langcode, langname, adminlang)
        VALUES ('%s', '%s', 'False')
        """ % (code, name)
        self.cursor.execute(query)
        self.connection.commit()

    def add_entry(self, tlang, source, target):
        """Add a entry to the DataBase."""
        
        # Precondition
        tlang = self._get_langcode(tlang)
        source = escape_msg(source)
        target = escape_msg(target)
        msgid = self._get_msgid(source)
        if not msgid:
            msgid = self._add_source(source)
        if self.search_entry(tlang, source):
            self._update_target(msgid, tlang, target)
        else:
            self._add_target(msgid, tlang, target)  
    
    def _add_source(self, source):
        """Add source to the DataBase. Return the new ID"""
        
        msgid = self._new_msgid()
        query = """
        INSERT INTO source(msgid, msg)
        VALUES (%d, '%s')
        """ % (msgid, source)
        self.cursor.execute(query)
        self.connection.commit()
        return msgid
    
    def _add_target(self, msgid, tlang, target):
        """Add target to the DataBase."""
        
        # Precondition
        msgid = int(msgid)
        query = """
        INSERT INTO target(msgid, langcode, tmsg)
        VALUES (%d, '%s', '%s')
        """ % (msgid, tlang, target)
        self.cursor.execute(query)
        self.connection.commit()

    def _update_target(self, msgid, tlang, target):
        """Update a target in the DataBase."""
        
        # Precondition
        msgid = int(msgid)
        query = """
        UPDATE target SET tmsg='%s'
        WHERE msgid=%d AND langcode='%s'
        """ % (target, msgid, tlang)
        self.cursor.execute(query)
        self.connection.commit()

    def _new_msgid(self):   
        """Get a new ID for a new source entry"""

        query = """
        SELECT msgid FROM source
        """
        self.cursor.execute(query)
        tmp = self.cursor.fetchall()
        msgid = len(tmp) + 1
        return msgid
        
class TM(TMWriter):

    """
    Translation Memory DataBase Interface.
    
    Usage example:
     obj = TM(fileName)
     source = "This is a entry to lookup"
     target = obj.get_entry(source, "Spanish")
     source = "This is a %"
     target = obj.match_entry(source, "ru")
     obj.close() 
    """
    
    def __init__(self, fileName):
        """Define dataFile."""
        
        self.connection = connect(fileName)
        self.cursor = self.connection.cursor()
        #TMReader.__init__(self, self.connection, self.cursor)
        TMWriter.__init__(self, self.connection, self.cursor)
        
    def close(self):
        """Close cursor and conection."""

        self.cursor.close()
        self.connection.close()

def create_tm(fileName):
    """Create the schema for the translation memory DataBase.""" 

    connection = connect(fileName)
    cursor = connection.cursor()
    
    # LANGUAGES:                                   
    # +----------+----------+-----------+
    # | langcode | langname | adminlang |
    # +----------+----------+-----------+
    query = """CREATE TABLE languages(langcode varchar(8) PRIMARY KEY, 
                                      langname varchar(32),
                                      adminlang boolean)"""
    cursor.execute(query)
    
    # SOURCE:                                  
    # +-------+------+
    # | msgid | term |
    # +-------+------+
    query = """CREATE TABLE source(msgid INTEGER PRIMARY KEY, 
                                   msg text)"""
    cursor.execute(query)

    # TARGET:                                  
    # +-------+----------+------+
    # | msgid | langcode | tmsg |
    # +-------+----------+------+
    query = """CREATE TABLE target(msgid int REFERENCES terms, 
                                   langcode varchar(8) REFERENCES languages, 
                                   tmsg text)"""
    cursor.execute(query)

    connection.commit()
    
    cursor.close()
    connection.close()

def escape_msg(msg):
    """Return string with special characters escaped."""
    return msg.replace("'", "''")

# EXCEPTIONS
# ==========
# The standard LookupError is used in most instances. 
class SlangDoesNotExist(Exception):
    pass

# TEST
# ====
import unittest
import os, os.path

class DbTest(unittest.TestCase):

    def setUp(self):
        self.NAME = 'TMTest'
        
        self.adminLang = {'code': 'en',
                          'name': 'English'}
        self.languages = []
        lang = {'code': 'es', 
                'name': 'Spanish'}
        self.languages.append(lang)
        lang = {'code': 'ru', 
                'name': 'Russian'}
        self.languages.append(lang)

        self.data = []
        tmp = {'tlang': 'es',
               'source': 'This is a simple entry',
               'target': 'Esta es una entrada sencilla'}
        self.data.append(tmp)   
        tmp = {'tlang': 'es',
               'source': 'This is a <b>second</b> entry',
               'target': 'Esta es una <b>segunda</b> entrada'}
        self.data.append(tmp)      
        tmp = {'tlang': 'es',
               'source': 'This is a <b>very <i>complex</i></b> entry',
               'target': 'Esta es una entrada <b>más <i>compleja</i></b>'}
        self.data.append(tmp)
        
    def tearDown(self):
        if os.path.isfile(self.NAME):
            os.unlink(self.NAME)
        
    def test_01(self):
        """ 1. Testing create_tm"""
    
        create_tm(self.NAME)
        self.failUnless(os.path.isfile(self.NAME), 'Database does not exist')

    def test_02(self):
        """2. Testing TMWriter"""
        
        create_tm(self.NAME)
        obj = TM(self.NAME)
        obj.add_slang(self.adminLang['name'], self.adminLang['code'])
        for lang in self.languages:
            obj.add_tlang(lang['name'], lang['code'])
            query = obj.get_tlangs()
            self.failUnless(query != None, 'Language was not added')
        for data in self.data:
            obj.add_entry(data['tlang'], data['source'], data['target'])
            query = obj.search_entry(data['tlang'], data['source'])
            self.failUnless(query != None, 'Entry was not added')
        self.failUnless(os.path.isfile(self.NAME), 'Database does not exist')
        obj.close()
    
    def test_03(self):
        """3. Testing TMReader"""
        
        create_tm(self.NAME)
        obj = TM(self.NAME)
        obj.add_slang(self.adminLang['name'], self.adminLang['code'])
        for lang in self.languages:
            obj.add_tlang(lang['name'], lang['code'])
        for data in self.data:
            obj.add_entry(data['tlang'], data['source'], data['target'])
        query = obj.search_entry('es', 'This is a simple entry')
        print query
        self.failUnless(query != None, 'First query failed')
        query = obj.match_entry('es', 'This is a % entry')
        print query
        self.failUnless(query != None, 'Second query failed')
        query = obj.match_entry('Spanish', 'This is a %complex% entry')
        print query[0].decode("utf8")
        self.failUnless(query != None, 'Third query failed')
        obj.close()

if __name__ == '__main__':
    unittest.main()
