### Copyright (C) 2006-2007 Manuel Ospina <mospina@redhat.com>

# This file is part of miaow.
#
# miaow is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# miaow is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with miaow; if not, write to the Free Software
# Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA

from tm import TM
import re
import difflib

class searchTM:

    """
    Query the database for exact or fuzzy matches.

    Methods:

     __init__(dataBase)
     __call__(tlang, source)
    """
    
    SOURCE = ''
    TLANG = ''
    SUGGEST = []
    TARGET = []

    def __init__(self, dataBase):
        """Connect to the database"""
        self.tm = TM(dataBase)
        
    def __call__(self, tlang, source):
        """Return exact or fuzzy  match of a string in a given language.
        
        'source' is the string to look for.
        'tlang' is the target language.

        The return value is a list with three elements:
        1. Rate: 1.0 is a exact match.
        2. Differences.
        3. Suggestion.
        """
        
        self.SOURCE = source
        self.TLANG = tlang
        result = self.tm.search_entry(self.TLANG, self.SOURCE)
        if result:
            return ['1.0', None, result]
        else:
            return self._do_metamorphosis()
    
    def _do_metamorphosis(self):
        """Transform source and query the database."""
        
        stage = 1
        while stage < 5:
            newSource = self._do_transformation(stage)
            for source in newSource:
                #print "Source:", source
                result = self.tm.match_entry(self.TLANG, source)
                #print "result:", result
                if result:
                    self.SUGGEST.extend(result)
            if self.SUGGEST:
                break
            stage += 1
        if not self.SUGGEST:
            return None
        return self._get_closest()
    
    def _do_transformation(self, stage):
        """Return a list of strings result of transforming self.SOURCE."""
        #print "STAGE:", stage
        # First transformation: <tags> and &entities;
        trans_1 = re.sub('<.+?>', '%', self.SOURCE)
        trans_1 = re.sub('&.+?;', '%', trans_1)
        if stage == 1:
            return [trans_1]
        # Second transformation: included
        trans_2 = [trans_1 + '%', '%' + trans_1]
        if stage == 2:
            return trans_2
        # Second transformation: sentences
        trans_3 = [ x.strip() for x in trans_1.split(".")]
        trans_3 = filter(lambda x: len(x) > 0, trans_3)
        if stage == 3:
            return trans_3 
        # Third transformation: included sentence
        trans_4 = map(lambda x: '% ' + x + ' %', trans_3)
        return trans_4

    def _get_closest(self):
        """Make a diff between suggestions and source to get the closest
        match."""
        
        s1 = self.SOURCE
        # Matching
        d = {}
        for s in self.SUGGEST:
            c = difflib.SequenceMatcher(lambda x: x == " ", s1, s)
            d[c.quick_ratio()] = [c, s]

        # Sorting (getting the closest match)
        p = None
        for i in d.keys():
            if not p or i == p:
                p = i
                continue
            elif i < p:
                continue
            else:
                p = i

        result = [str(p)]
        diffs = ''
        # Getting differences
        for block in d[p][0].get_opcodes():
            if block[0] == 'equal':
                continue
            if block[0] == 'delete':
                diffs += '-' + s1[block[1]:block[2]] + '\n'
            if block[0] == 'insert':
                diffs += '+' + s1[block[1]:block[2]] + '\n'
            if block[0] == 'replace':
                diffs += '*' + s1[block[1]:block[2]] + '\n'
        result.append(diffs)        
        result.append(d[p][1].strip())
        return result       

# [TEST]
# [TODO]
# Implement test cases.
if __name__ == '__main__':
    import sys
    t = searchTM('testTM')
    r = t('es', sys.argv[1])
    print r
